# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class EditableCloudResource(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'aws_s3_id': 'str',
        'aws_efs_id': 'str',
        'aws_efs_mount_target_ip': 'str',
        'memorydb_cluster_config': 'AWSMemoryDBClusterConfig'
    }

    attribute_map = {
        'aws_s3_id': 'aws_s3_id',
        'aws_efs_id': 'aws_efs_id',
        'aws_efs_mount_target_ip': 'aws_efs_mount_target_ip',
        'memorydb_cluster_config': 'memorydb_cluster_config'
    }

    def __init__(self, aws_s3_id=None, aws_efs_id=None, aws_efs_mount_target_ip=None, memorydb_cluster_config=None, local_vars_configuration=None):  # noqa: E501
        """EditableCloudResource - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._aws_s3_id = None
        self._aws_efs_id = None
        self._aws_efs_mount_target_ip = None
        self._memorydb_cluster_config = None
        self.discriminator = None

        if aws_s3_id is not None:
            self.aws_s3_id = aws_s3_id
        if aws_efs_id is not None:
            self.aws_efs_id = aws_efs_id
        if aws_efs_mount_target_ip is not None:
            self.aws_efs_mount_target_ip = aws_efs_mount_target_ip
        if memorydb_cluster_config is not None:
            self.memorydb_cluster_config = memorydb_cluster_config

    @property
    def aws_s3_id(self):
        """Gets the aws_s3_id of this EditableCloudResource.  # noqa: E501

        The AWS S3 id of this cloud.  # noqa: E501

        :return: The aws_s3_id of this EditableCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._aws_s3_id

    @aws_s3_id.setter
    def aws_s3_id(self, aws_s3_id):
        """Sets the aws_s3_id of this EditableCloudResource.

        The AWS S3 id of this cloud.  # noqa: E501

        :param aws_s3_id: The aws_s3_id of this EditableCloudResource.  # noqa: E501
        :type: str
        """

        self._aws_s3_id = aws_s3_id

    @property
    def aws_efs_id(self):
        """Gets the aws_efs_id of this EditableCloudResource.  # noqa: E501

        The AWS EFS id of this cloud.  # noqa: E501

        :return: The aws_efs_id of this EditableCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._aws_efs_id

    @aws_efs_id.setter
    def aws_efs_id(self, aws_efs_id):
        """Sets the aws_efs_id of this EditableCloudResource.

        The AWS EFS id of this cloud.  # noqa: E501

        :param aws_efs_id: The aws_efs_id of this EditableCloudResource.  # noqa: E501
        :type: str
        """

        self._aws_efs_id = aws_efs_id

    @property
    def aws_efs_mount_target_ip(self):
        """Gets the aws_efs_mount_target_ip of this EditableCloudResource.  # noqa: E501

        The AWS EFS mount target IP of this cloud.  # noqa: E501

        :return: The aws_efs_mount_target_ip of this EditableCloudResource.  # noqa: E501
        :rtype: str
        """
        return self._aws_efs_mount_target_ip

    @aws_efs_mount_target_ip.setter
    def aws_efs_mount_target_ip(self, aws_efs_mount_target_ip):
        """Sets the aws_efs_mount_target_ip of this EditableCloudResource.

        The AWS EFS mount target IP of this cloud.  # noqa: E501

        :param aws_efs_mount_target_ip: The aws_efs_mount_target_ip of this EditableCloudResource.  # noqa: E501
        :type: str
        """

        self._aws_efs_mount_target_ip = aws_efs_mount_target_ip

    @property
    def memorydb_cluster_config(self):
        """Gets the memorydb_cluster_config of this EditableCloudResource.  # noqa: E501

        The AWS memorydb cluster config of this cloud.  # noqa: E501

        :return: The memorydb_cluster_config of this EditableCloudResource.  # noqa: E501
        :rtype: AWSMemoryDBClusterConfig
        """
        return self._memorydb_cluster_config

    @memorydb_cluster_config.setter
    def memorydb_cluster_config(self, memorydb_cluster_config):
        """Sets the memorydb_cluster_config of this EditableCloudResource.

        The AWS memorydb cluster config of this cloud.  # noqa: E501

        :param memorydb_cluster_config: The memorydb_cluster_config of this EditableCloudResource.  # noqa: E501
        :type: AWSMemoryDBClusterConfig
        """

        self._memorydb_cluster_config = memorydb_cluster_config

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, EditableCloudResource):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, EditableCloudResource):
            return True

        return self.to_dict() != other.to_dict()
